from fmtr.tools import env, Path, net
from haco.paths import paths

STATUS = 'status'
HOSTNAME = net.get_hostname().lower()
F = net.get_fqdn()
MAC = net.MAC()

CLIENT_ID = f'{HOSTNAME}-{MAC.hex}'

TOPIC_ROOT = Path(paths.name) / env.CHANNEL
TOPIC_CLIENT = TOPIC_ROOT / CLIENT_ID
TOPIC_AVAIL = TOPIC_CLIENT / STATUS
MQTT_HOST = "mqtt.service"
MQTT_PORT = 1883
MQTT_USERNAME = "user"
MQTT_PASSWORD = None



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Type

from fmtr.tools.iterator_tools import strip_none
from haco.base import Base
from haco.topics import AnnounceTopicState, AnnounceTopicCommand
from haco.utils import Converters

if TYPE_CHECKING:
    from haco.control import Control


@dataclass(kw_only=True)
class Capability(Base):
    name: str | None = None

    converters: Type[Converters] | None = field(default=None, metadata=dict(exclude=True))

    state: AnnounceTopicState | None = field(default_factory=AnnounceTopicState)
    command: AnnounceTopicCommand | None = field(default_factory=AnnounceTopicCommand)


    parent: Control | None = None
    announce = None

    def set_parent(self, control):
        self.parent = control
        for topic in self.topics:
            topic.set_parent(self)

        self.converters = self.converters or self.control.converters

    def get_announce(self) -> dict:
        data = {}
        for topic in self.topics:
            data |= topic.get_announce()
        return data

    def get_subscriptions(self) -> dict:
        data = {}
        for topic in self.topics:
            data |= topic.get_subscriptions()
        return data

    @property
    def topics(self):
        return strip_none(self.state, self.command)

    @property
    def control(self):
        return self.parent



from __future__ import annotations

from dataclasses import dataclass, fields
from typing import Any


@dataclass(kw_only=True)
class Base:
    DATA = dict()

    def model_dump(self) -> dict[str, Any]:
        data = {} | self.DATA
        for fld in fields(self):
            meta = fld.metadata
            if meta.get('exclude', False):
                continue

            value = getattr(self, fld.name)

            if hasattr(value, self.model_dump.__name__):
                value = value.model_dump()
            elif isinstance(value, list):
                value = [v.model_dump() if hasattr(v, self.model_dump.__name__) else v for v in value]

            if value is not None:
                data[fld.name] = value
        return data



from dataclasses import dataclass
from typing import Literal

from haco.capabilities import Capability
from haco.control import Control


@dataclass(kw_only=True)
class Climate(Control):
    DATA = dict(
        platform='climate'
    )


    temperature_unit: Literal["C", "F"] = "C"
    temperature_min: float = 0
    temperature_max: float = 40
    humidity_min: float = 0
    humidity_max: float = 100
    precision: int = 1
    modes: list[str] | None = None
    preset_modes: list[str] | None = None
    fan_modes: list[str] | None = None
    swing_modes: list[str] | None = None


    def temperature_command(self, value):
        raise NotImplementedError()

    def mode_command(self, value):
        raise NotImplementedError()

    def fan_mode_command(self, value):
        raise NotImplementedError()

    def fan_mode_state(self, value):
        raise NotImplementedError()

    def power_command(self, value):
        raise NotImplementedError()

    def power_state(self, value):
        raise NotImplementedError()

    @classmethod
    def get_capabilities(cls):
        return [
            Capability(name="current_temperature", command=None),
            Capability(name="temperature"),
            Capability(name="current_humidity"),
            Capability(name="target_humidity"),
            Capability(name="mode"),
            Capability(name="fan_mode"),
            Capability(name="power"),
        ]



import asyncio
import random

import aiomqtt

from fmtr.tools import Constants, mqtt
from fmtr.tools.json_tools import to_json
from haco import constants


async def handle_commands(client, control):
    """Listen for command messages from HA."""

    subs = control.subscriptions

    async with client.messages() as messages:

        for topic_sub in subs.keys():
            await client.subscribe(topic_sub)

        async for message in messages:
            payload = message.payload.decode()
            print(f"{message.topic.value}{Constants.ARROW}{payload}")
            method = subs[message.topic.value]
            echo_val = method(payload)
            echo_val

            # Echo back as new state
            # await client.publish(STATE_TOPIC, payload, retain=True)


async def background_task(client, control):
    while True:
        await asyncio.sleep(10)
        cap = control.capabilities[1]
        value = random.randint(18, 24)
        print(f'human is setting temperature to {value}')
        await client.publish(cap.state.topic, payload=to_json(value), retain=True)


class ClientHaco(mqtt.Client):

    def __init__(self, *args, device=None, will=None, client_id=None, **kwargs):
        self.will = will or mqtt.Will(topic=str(constants.TOPIC_AVAIL), payload="offline", retain=True, qos=1, )
        client_id = client_id or constants.CLIENT_ID
        self.device = device
        # self.device.client = self
        super().__init__(*args, will=self.will, client_id=client_id, **kwargs)


async def main():
    # Define capabilities

    # device = Device(
    #     name="Development Device",
    #     controls=[
    #         Climate(
    #             #id="climate_living_room",
    #             name="Climate Control",
    #             fan_modes=['a','b','c'],
    #         )
    #     ]
    #
    # )

    # announce=device.announce
    # subs=device.subscriptions

    MQTT_HOST = "mqtt.service"
    MQTT_PORT = 1883
    MQTT_USERNAME = "user"
    MQTT_PASSWORD = None

    while True:
        try:
            async with ClientHaco(
                    hostname=MQTT_HOST,
                    port=MQTT_PORT,
                    username=MQTT_USERNAME,
                    password=MQTT_PASSWORD,

                    client_id=constants.CLIENT_ID,
                    # device=device,
            ) as client:

                # for topic, data in announce.items():
                #     data_json = to_json(data)
                #     print(f'Announcing {topic} with {data_json}')
                #     await client.publish(topic, payload=data_json, retain=True)

                # Online now
                await client.publish(client.will.topic, "online", retain=True)

                # await publish_discovery(client)
                # await client.publish(STATE_TOPIC, "OFF", retain=True)

                # Wait for commands
                # asyncio.create_task(background_task(client,control=climate))
                # await handle_commands(client,control=climate)
                client

        except aiomqtt.MqttError as e:
            print(f"⚠️ MQTT disconnected: {e}")
            # await asyncio.sleep(5)


if __name__ == "__main__":
    asyncio.run(main())



from enum import StrEnum


class Uom(StrEnum):
    """

    Sourced from: https://github.com/home-assistant/core/blob/03b1c7ad1deb4b9494df01b27ef8e1d7dfe179ba/homeassistant/const.py#L489

    """
    POWER_WATT = "W"
    POWER_KILO_WATT = "kW"

    VOLT = "V"

    ENERGY_WATT_HOUR = "Wh"
    ENERGY_KILO_WATT_HOUR = "kWh"

    ELECTRICAL_CURRENT_AMPERE = "A"
    ELECTRICAL_VOLT_AMPERE = "VA"

    DEGREE = "°"

    CURRENCY_EURO = "€"
    CURRENCY_DOLLAR = "$"
    CURRENCY_CENT = "¢"

    TEMP_CELSIUS = "°C"
    TEMP_FAHRENHEIT = "°F"
    TEMP_KELVIN = "K"

    TIME_MICROSECONDS = "μs"
    TIME_MILLISECONDS = "ms"
    TIME_SECONDS = "s"
    TIME_MINUTES = "min"
    TIME_HOURS = "h"
    TIME_DAYS = "d"
    TIME_WEEKS = "w"
    TIME_MONTHS = "m"
    TIME_YEARS = "y"

    LENGTH_MILLIMETERS = "mm"
    LENGTH_CENTIMETERS = "cm"
    LENGTH_METERS = "m"
    LENGTH_KILOMETERS = "km"

    LENGTH_INCHES = "in"
    LENGTH_FEET = "ft"
    LENGTH_YARD = "yd"
    LENGTH_MILES = "mi"

    FREQUENCY_HERTZ = "Hz"
    FREQUENCY_GIGAHERTZ = "GHz"

    PRESSURE_PA = "Pa"
    PRESSURE_HPA = "hPa"
    PRESSURE_BAR = "bar"
    PRESSURE_MBAR = "mbar"
    PRESSURE_INHG = "inHg"
    PRESSURE_PSI = "psi"

    VOLUME_LITERS = "L"
    VOLUME_MILLILITERS = "mL"
    VOLUME_CUBIC_METERS = "m³"
    VOLUME_CUBIC_FEET = "ft³"

    VOLUME_GALLONS = "gal"
    VOLUME_FLUID_OUNCE = "fl. oz."

    VOLUME_FLOW_RATE_CUBIC_METERS_PER_HOUR = "m³/h"
    VOLUME_FLOW_RATE_CUBIC_FEET_PER_MINUTE = "ft³/m"

    AREA_SQUARE_METERS = "m²"

    MASS_GRAMS = "g"
    MASS_KILOGRAMS = "kg"
    MASS_MILLIGRAMS = "mg"
    MASS_MICROGRAMS = "µg"

    MASS_OUNCES = "oz"
    MASS_POUNDS = "lb"

    CONDUCTIVITY = "µS/cm"

    LIGHT_LUX = "lx"

    UV_INDEX = "UV index"

    PERCENTAGE = "%"

    IRRADIATION_WATTS_PER_SQUARE_METER = "W/m²"

    PRECIPITATION_MILLIMETERS_PER_HOUR = "mm/h"

    CONCENTRATION_MICROGRAMS_PER_CUBIC_METER = "µg/m³"
    CONCENTRATION_MILLIGRAMS_PER_CUBIC_METER = "mg/m³"
    CONCENTRATION_PARTS_PER_CUBIC_METER = "p/m³"
    CONCENTRATION_PARTS_PER_MILLION = "ppm"
    CONCENTRATION_PARTS_PER_BILLION = "ppb"

    SPEED_MILLIMETERS_PER_DAY = "mm/d"
    SPEED_INCHES_PER_DAY = "in/d"
    SPEED_METERS_PER_SECOND = "m/s"
    SPEED_INCHES_PER_HOUR = "in/h"
    SPEED_KILOMETERS_PER_HOUR = "km/h"
    SPEED_MILES_PER_HOUR = "mph"

    SIGNAL_STRENGTH_DECIBELS = "dB"
    SIGNAL_STRENGTH_DECIBELS_MILLIWATT = "dBm"

    DATA_BITS = "bit"
    DATA_KILOBITS = "kbit"
    DATA_MEGABITS = "Mbit"
    DATA_GIGABITS = "Gbit"
    DATA_BYTES = "B"
    DATA_KILOBYTES = "kB"
    DATA_MEGABYTES = "MB"
    DATA_GIGABYTES = "GB"
    DATA_TERABYTES = "TB"
    DATA_PETABYTES = "PB"
    DATA_EXABYTES = "EB"
    DATA_ZETTABYTES = "ZB"
    DATA_YOTTABYTES = "YB"
    DATA_KIBIBYTES = "KiB"
    DATA_MEBIBYTES = "MiB"
    DATA_GIBIBYTES = "GiB"
    DATA_TEBIBYTES = "TiB"
    DATA_PEBIBYTES = "PiB"
    DATA_EXBIBYTES = "EiB"
    DATA_ZEBIBYTES = "ZiB"
    DATA_YOBIBYTES = "YiB"

    DATA_RATE_BITS_PER_SECOND = "bit/s"
    DATA_RATE_KILOBITS_PER_SECOND = "kbit/s"
    DATA_RATE_MEGABITS_PER_SECOND = "Mbit/s"
    DATA_RATE_GIGABITS_PER_SECOND = "Gbit/s"
    DATA_RATE_BYTES_PER_SECOND = "B/s"
    DATA_RATE_KILOBYTES_PER_SECOND = "kB/s"
    DATA_RATE_MEGABYTES_PER_SECOND = "MB/s"
    DATA_RATE_GIGABYTES_PER_SECOND = "GB/s"
    DATA_RATE_KIBIBYTES_PER_SECOND = "KiB/s"
    DATA_RATE_MEBIBYTES_PER_SECOND = "MiB/s"
    DATA_RATE_GIBIBYTES_PER_SECOND = "GiB/s"



import aiomqtt
import asyncio

from fmtr.tools import mqtt, Constants
from haco import constants
from haco.obs import logger


class ClientHaco(mqtt.Client):
    ONLINE = 'online'
    OFFLINE = 'offline'

    aiomqtt.Client

    def __init__(self, *args, device=None, will=None, identifier=None, **kwargs):
        self.will = will or mqtt.Will(topic=str(constants.TOPIC_AVAIL), payload=self.OFFLINE, retain=True, qos=1, )
        identifier = identifier or constants.CLIENT_ID
        self.device = device
        self.device.set_parent(self)
        super().__init__(*args, will=self.will, identifier=identifier, **kwargs)

    @property
    def topic(self):
        return constants.TOPIC_CLIENT

    async def handle(self):
        """Listen for command messages from HA."""        

        for topic_sub in self.device.subscriptions.keys():
            logger.info(f"Subscribing to {topic_sub}")
            await self.subscribe(topic_sub)

        async for message in self.messages:
            payload = message.payload.decode()
            logger.info(f"{message.topic.value}{Constants.ARROW}{payload}")
            topic_command = self.device.subscriptions[message.topic.value]
            echo_val = await topic_command.handle(message)



    async def start(self):
        while True:
            try:
                async with self:

                    await self.device.announce()

                    await self.publish(self.will.topic, self.ONLINE, retain=True)

                    # await self.publish('haco/development/3012edb1a6d4-1ed9c3469373/dev-device/dev-climate-b/power/command', "OFF", retain=True)

                    await self.handle()

            except aiomqtt.MqttError as e:
                logger.info(f"⚠️ MQTT disconnected: {e}; retrying in 5s")
                await asyncio.sleep(5)



import string
from dataclasses import dataclass
from typing import Callable

from fmtr.tools import json
from fmtr.tools.datatype_tools import to_bool

CHARS_ALLOWED = string.ascii_lowercase + string.digits
SEPS = '_- /'
BRANCH = 'release'

CAPABILITY_DEFAULT = 'default'
RECONNECT_DELAY_SEC = 5
ON = 'ON'
OFF = 'OFF'

BOOL_MAP = {True: ON, False: OFF}

from_bool = BOOL_MAP.get

def sanitize_name(s, sep='-'):
    chars = []

    for c in s.lower():
        if c in SEPS:
            chars.append(sep)
        elif c in CHARS_ALLOWED:
            chars.append(c)

    sanitized = ''.join(chars)

    if not sanitized:
        raise ValueError("Sanitized string is empty")

    return sanitized


class Converters:
    command: Callable = None
    state: Callable = None


class ConvertersBool(Converters):
    command: Callable = to_bool
    state: Callable = from_bool


class ConvertersString(Converters):
    command: Callable = str
    state: Callable = str


class ConvertersNumeric(Converters):
    command: Callable = json.from_json
    state: Callable = json.to_json

@dataclass
class Metadata:
    exclude: bool = False



import asyncio
import json

import aiomqtt

from haco.constants import MQTT_HOST, MQTT_PORT, MQTT_USERNAME, MQTT_PASSWORD

DEVICE_ID = "haco_sandbox"
ENTITY_ID = "demo_switch"
DEVICE_NAME = "Haco Sandbox Device"
ENTITY_NAME = "Demo Switch"

# Topics
CONFIG_TOPIC = f"homeassistant/switch/{ENTITY_ID}/config"
STATE_TOPIC = f"haco/{DEVICE_ID}/switch/state"
COMMAND_TOPIC = f"haco/{DEVICE_ID}/switch/set"
AVAIL_TOPIC = f"haco/{DEVICE_ID}/status"


async def publish_discovery(client):
    """Announce the entity to Home Assistant (retained)."""
    config = {
        "name": ENTITY_NAME,
        "unique_id": ENTITY_ID,
        "state_topic": STATE_TOPIC,
        "command_topic": COMMAND_TOPIC,
        "availability_topic": AVAIL_TOPIC,
        "payload_on": "ON",
        "payload_off": "OFF",
        "payload_available": "online",
        "payload_not_available": "offline",
        "device": {
            "identifiers": [DEVICE_ID],
            "name": DEVICE_NAME,
            "manufacturer": "Demo",
            "model": "Python MQTT Example"
        }
    }

    await client.publish(CONFIG_TOPIC, json.dumps(config), retain=True)
    print(f"📢 Announced entity via {CONFIG_TOPIC}")


async def handle_commands(client):
    """Listen for command messages from HA."""
    async with client.messages() as messages:
        await client.subscribe(COMMAND_TOPIC)
        async for message in messages:
            payload = message.payload.decode()
            print(f"📥 Received command: {payload}")

            # Echo back as new state
            await client.publish(STATE_TOPIC, payload, retain=True)


async def main():
    while True:
        try:
            async with aiomqtt.Client(
                    hostname=MQTT_HOST,
                    port=MQTT_PORT,
                    username=MQTT_USERNAME,
                    password=MQTT_PASSWORD,
                    will=aiomqtt.Will(
                        topic=AVAIL_TOPIC,
                        payload="offline",
                        retain=True,
                        qos=1,
                    ),
            ) as client:

                # Online now
                await client.publish(AVAIL_TOPIC, "online", retain=True)
                await publish_discovery(client)
                await client.publish(STATE_TOPIC, "OFF", retain=True)

                # Wait for commands
                await handle_commands(client)

        except aiomqtt.MqttError as e:
            print(f"⚠️ MQTT disconnected: {e}; retrying in 5s")
            await asyncio.sleep(5)


if __name__ == "__main__":
    asyncio.run(main())
